import os
import sys

from pyocd.target.pack.pack_target import PackTargets

from trackpad_configuration_manager.utils.execUtils import resource_path


class CMSISPackage:

    @staticmethod
    def load_package():
        # Load CMSIS-package for stm32f4xx

        stm32f0_pack_path = None
        stm32u5_pack_path = None
        stm32g0_pack_path = None

        # determine if application is a script file or frozen exe
        if getattr(sys, 'frozen', False) and hasattr(sys, '_MEIPASS'):
            stm32f0_pack_path = resource_path("trackpad_configuration_manager/cmsis-pack/Keil.STM32F0xx_DFP.2.1.1.pack")
            stm32u5_pack_path = resource_path("trackpad_configuration_manager/cmsis-pack/Keil.STM32U5xx_DFP.2.0.0.pack")
            stm32g0_pack_path = resource_path("trackpad_configuration_manager/cmsis-pack/Keil.STM32G0xx_DFP.1.4.0.pack")
        else:
            dir_name = os.path.dirname(__file__)
            stm32f0_pack_path = os.path.join(dir_name, "cmsis-pack/Keil.STM32F0xx_DFP.2.1.1.pack")
            stm32u5_pack_path = os.path.join(dir_name, "cmsis-pack/Keil.STM32U5xx_DFP.2.0.0.pack")
            stm32g0_pack_path = os.path.join(dir_name, "cmsis-pack/Keil.STM32G0xx_DFP.1.4.0.pack")

        PackTargets.populate_targets_from_pack(
            [stm32f0_pack_path, stm32u5_pack_path, stm32g0_pack_path])



from trackpad_configuration_manager.dataHolder import DataHolder
from trackpad_configuration_manager.utils.crc16 import compute_crc


def process_struct(element: dict) -> bytearray():
    data = bytearray()
    for key, value in element.items():
        data += process_type(value)

    return data


def process_value(data_holder: DataHolder) -> bytearray():
    data = None

    if isinstance(data_holder, DataHolder):
        data = data_holder.to_bytes()
    else:
        data = bytes()

    return bytearray(data)


def process_value_array_proxy(element) -> bytearray():
    data = bytearray()
    if isinstance(element, list):
        for value in element:
            data += process_value(value)
    else:
        data += process_value(element)
    return data


def process_value_array(array_value: list, _type: str) -> bytearray():
    data = bytearray()
    for value in array_value:
        match _type:
            case "struct":
                data += process_struct(value)
            case _:
                data += process_value_array_proxy(value)
    return data


def process_type(element: dict) -> bytearray():
    data = bytearray()
    if "type" in element and "value" in element:
        if isinstance(element["value"], DataHolder):
            data = process_value(element["value"])
        elif isinstance(element["value"], dict):
            data = process_struct(element["value"])
        elif isinstance(element["value"], list):
            data = process_value_array(element["value"], element["type"])

    return data


def generate_binary(dict_json: dict) -> bytearray():
    data = bytearray()
    for key, value in dict_json.items():
        data += process_type(value)

    return data


class ConfMarshaller:
    def __init__(self, json_head_struct: dict):

        if "version" not in json_head_struct:
            raise Exception("Json dict doesn't contains mandatory version element")

        self.data_on_file = json_head_struct

    def get_binary_for_static_memory(self) -> [bytearray(), int]:
        data = generate_binary(self.data_on_file)

        checksum = compute_crc(data)
        crc_byte = checksum.to_bytes(length=2, byteorder="little", signed=False)

        return bytearray(crc_byte) + data, checksum

    def __get_all_dynamic_data_set(self, d) -> list:
        list_data_set = list()

        if isinstance(d, DataHolder):
            return [d.get_dynamic_memory_section()]
        if not isinstance(d, (dict, list)):
            return list()
        if isinstance(d, list):
            for v in d:
                list_data_set += self.__get_all_dynamic_data_set(v)
            return list_data_set

        list_data_set = list()
        for k, v in d.items():
            list_data_set += self.__get_all_dynamic_data_set(v)

        return list_data_set

    def get_binaries_for_dynamic_memory(self) -> bytearray():

        list_data_set = self.__get_all_dynamic_data_set(self.data_on_file)

        data = bytearray()
        list_data_set = list(filter(None, list_data_set))

        data = len(list_data_set).to_bytes(length=1, byteorder="little", signed=False)

        for data_set in list_data_set:
            data += bytearray(data_set.to_bytes())

        return data



import io
import logging
import os
import shutil
from abc import abstractmethod
from enum import Enum

from PySide6.QtCore import Signal, QObject

from trackpad_configuration_manager.utils.crc16 import compute_crc


class DynamicDataSetType(Enum):
    WAVEFILE = 1
    BOS1921_REFERENCED_FEEDBACK = 2
    FILE_SYSTEM_FILE = 3


class DynamicDataSet:
    def __init__(self, data: bytes, data_type: DynamicDataSetType):
        self.data = data
        self.type = data_type

    def to_bytes(self) -> bytes:
        data = bytearray(self.type.value.to_bytes(length=1, byteorder="little", signed=False))  # data type
        data += bytearray(len(self.data).to_bytes(length=2, byteorder="little", signed=False))  # data length
        data += bytearray(self.data)  # data

        data += bytearray(compute_crc(data).to_bytes(length=2, byteorder="little", signed=False))  # crc

        return bytes(data)


class WavefileDataSet(DynamicDataSet):

    def __init__(self, data):
        super().__init__(data, DynamicDataSetType.WAVEFILE)


class FileSystemFileDataSet(DynamicDataSet):

    def __init__(self, data, device_file_path: str, path_length: int):
        super().__init__(data, DynamicDataSetType.FILE_SYSTEM_FILE)
        self.device_file_path = device_file_path
        self.path_length = path_length

    def to_bytes(self) -> bytes:
        data = bytearray(self.type.value.to_bytes(length=1, byteorder="little", signed=False))  # data type
        data += bytearray(
            len(self.device_file_path).to_bytes(length=2, byteorder="little", signed=False))  # Filepath length
        data += bytearray(self.device_file_path.encode())  # Filepath
        data += bytearray(len(self.data).to_bytes(length=2, byteorder="little", signed=False))  # data length
        data += bytearray(self.data)  # data
        data += bytearray(compute_crc(data).to_bytes(length=2, byteorder="little", signed=False))  # crc
        return bytes(data)


class DataHolder(QObject):
    updated_device_value = Signal(str, bool)

    def __init__(self):
        QObject.__init__(self)

    @abstractmethod
    def read_device_value(self, data: io.BytesIO):
        pass

    @abstractmethod
    def to_bytes(self) -> bytes:
        pass

    @abstractmethod
    def get_file_value(self):
        pass

    @abstractmethod
    def get_dynamic_memory_section(self) -> DynamicDataSet:
        pass


class BoolTypeDataHolder(DataHolder):
    def __init__(self, file_value: bool, is_read_only: bool = False):
        self._file_value = file_value
        self._device_value = 0
        self._device_bool = False
        self.length = 1
        self.is_read_only = is_read_only

        super().__init__()

    def read_device_value(self, data: io.BytesIO):
        self._device_value = int.from_bytes(data.read(self.length), byteorder="little", signed=False)

        if self._device_value == 1:
            self._device_bool = True
        elif self._device_value == 0:
            self._device_bool = False
        else:
            raise Exception("Invalid value")
        self.update_device_value_label()

    def to_bytes(self) -> bytes:
        value = 0

        if self._file_value:
            value = 1

        return value.to_bytes(length=self.length, byteorder="little", signed=False)

    def get_file_value(self):
        return self._file_value

    def get_file_value_int(self):
        if self._file_value:
            return 1

        return 0

    def get_device_value(self):
        return self._device_bool

    def get_device_str_value(self):
        return str(self._device_bool)

    def get_dynamic_memory_section(self) -> DynamicDataSet:
        pass

    def set_file_value(self, value: int):
        if value == 1:
            self._file_value = True
        elif value == 0:
            self._file_value = False
        else:
            raise Exception("Invalid value")
        self.update_device_value_label()

    def update_device_value_label(self):
        self.updated_device_value.emit(str(self._device_bool), self._device_bool != self._file_value)


class NativeTypeDataHolder(DataHolder):
    SUPPORTED_TYPE = {"uint8_t": {"signed": False, "length": 1, "min": 0, "max": 255},
                      "uint16_t": {"signed": False, "length": 2, "min": 0, "max": 65536},
                      "uint32_t": {"signed": False, "length": 4, "min": 0, "max": 4294967296},
                      "int8_t": {"signed": True, "length": 1, "min": -127, "max": 128},
                      "int16_t": {"signed": True, "length": 2, "min": -32767, "max": 32768},
                      "int32_t": {"signed": True, "length": 4, "min": -2147483647, "max": 2147483648}}

    DISPLAY_TYPE = {"hex", "bin", "dec"}

    def __init__(self, file_value: int, type: str, is_read_only, choices: dict = None, display_type: str = "dec"):
        self.type = type
        self._file_value = file_value
        self._device_value = 0
        self.is_read_only = is_read_only
        self.is_multiple_choice = False
        self.choices = None

        if len(display_type) > 0:
            if display_type not in self.DISPLAY_TYPE:
                raise Exception("Invalid display type")
            self.display_type = display_type
        else:
            self.display_type = "dec"

        if choices:
            self.is_multiple_choice = True
            self.choices = choices

        if type not in self.SUPPORTED_TYPE:
            raise Exception("Invalid type")

        super().__init__()

    def set_file_value(self, value: str):
        __value = 0

        match self.display_type:
            case "hex":
                # validate this is a valid hex string
                if not value.startswith("0x") or not value[2:].isalnum():
                    raise Exception("Invalid hex value")
                __value = int(value, 16)
            case "bin":
                if len(value) < 0 and value.isdigit() is False:
                    raise Exception("Invalid value")
                __value = int(value, 2)
            case "dec":
                if len(value) < 0 and value.isdigit() is False:
                    raise Exception("Invalid value")
                __value = int(value, 10)

        if __value < self.SUPPORTED_TYPE[self.type]["min"] or __value > self.SUPPORTED_TYPE[self.type]["max"]:
            raise Exception("out of range value")

        self._file_value = __value
        self.update_device_value_label()

    def set_file_value_from_int(self, value: int):
        if value < self.SUPPORTED_TYPE[self.type]["min"] or value > self.SUPPORTED_TYPE[self.type]["max"]:
            raise Exception("out of range value")

        self._file_value = value
        self.update_device_value_label()

    def set_file_value_from_choice(self, value: str):
        if not self.is_multiple_choice:
            raise Exception("Instance not multiple choices")

        if value not in self.choices:
            raise Exception("Invalid value")

        self._file_value = list(self.choices.keys()).index(value)
        self.update_device_value_label()

    def get_file_value(self) -> int:
        return self._file_value

    def get_file_str_value(self) -> str:
        return self.__convert_value_str(self._file_value)

    def get_file_str_value_from_choice(self) -> str:
        if self.is_multiple_choice:
            for choice, value in self.choices.items():
                if value == self._file_value:
                    return choice

            raise Exception("No corresponding choice")
        raise Exception("Instance not multiple choices")

    def get_file_index_from_choice(self) -> int:
        if self.is_multiple_choice:
            index = 0
            for choice, value in self.choices.items():
                if value == self._file_value:
                    return index
                index += 1
            raise Exception("No corresponding choice")
        raise Exception("Instance not multiple choices")

    def get_device_value(self) -> int:
        return self._device_value

    def get_device_str_value_from_choice(self) -> str:
        if self.is_multiple_choice:
            for choice, value in self.choices.items():
                if value == self._device_value:
                    return choice

            raise Exception("No corresponding choice")
        raise Exception("Instance not multiple choices")

    def get_choices_str(self):
        str_choices = []

        for choice in self.choices:
            str_choices.append(choice)

        return str_choices

    def __convert_value_str(self, value: int) -> str:
        match self.display_type:
            case "hex":
                return "0x" + hex(value)[2:].upper().zfill(self.SUPPORTED_TYPE[self.type]["length"])
            case "bin":
                # return a string with 0b prefix
                return "0b" + bin(value)[2:].zfill(self.SUPPORTED_TYPE[self.type]["length"] * 8)
            case "dec":
                return str(value)

    def get_device_str_value(self) -> str:
        return self.__convert_value_str(self._device_value)

    def to_bytes(self) -> bytes:
        return self._file_value.to_bytes(length=self.SUPPORTED_TYPE[self.type]["length"], byteorder="little",
                                         signed=self.SUPPORTED_TYPE[self.type]["signed"])

    def read_device_value(self, data: io.BytesIO):
        length = self.SUPPORTED_TYPE[self.type]["length"]

        self._device_value = int.from_bytes(data.read(length), byteorder="little",
                                            signed=self.SUPPORTED_TYPE[self.type]["signed"])
        self.update_device_value_label()

    def get_dynamic_memory_section(self) -> DynamicDataSet:
        pass

    def update_device_value_label(self):
        str_device_value = self.get_device_str_value_from_choice() if self.is_multiple_choice else self.get_device_str_value()
        self.updated_device_value.emit(str_device_value, self._device_value != self._file_value)


class WaveFile:
    def __init__(self, name, path):
        self.name = name
        self.path = path


class DynamicSectionNotFoundException(Exception):
    def __init__(self, path: str):
        self.path = path
        super().__init__("Cannot find dynamic section at : " + path)


class WavefileDataHolder(DataHolder):

    def get_device_str_value(self):
        pass

    def __init__(self, file_value: dict, base_location: str):
        self.on_file = WaveFile(file_value["name"], file_value["local-path"])
        self.on_device = WaveFile(file_value["name"], file_value["local-path"])
        if file_value.get("length-name") is None:
            self.length_name = 0
        else:
            self.length_name = file_value["length-name"]
        if file_value.get("length-path") is None:
            self.length_path = 0
        else:
            self.length_path = file_value["length-path"]
        self.base_location = os.path.dirname(os.path.abspath(base_location))
        self.path_device = "ramfs:/" + self.on_file.name

        super().__init__()

    def update_on_file(self, file_path: str):

        if self.base_location not in file_path:
            # if there is already a file with the same name, remove it
            if os.path.exists(os.path.join(self.base_location, os.path.basename(file_path))):
                os.remove(os.path.join(self.base_location, os.path.basename(file_path)))

            # copy file to the base location in the resources folder
            shutil.copy(file_path, self.base_location)
            file_path = os.path.join(self.base_location, os.path.basename(file_path))

        self.on_file.path = os.path.relpath(file_path, self.base_location)
        self.on_file.name = os.path.basename(file_path)
        self.update_device_value_label()

    def get_full_path_on_file(self) -> str:
        return self.on_file.path

    def read_device_value(self, data: io.BytesIO):
        name_data = data.read(self.length_name)

        path_data = data.read(self.length_path)

        try:
            self.on_device.name = name_data.decode('ascii').strip().strip('\x00')
            self.on_device.path = path_data.decode('ascii').strip().strip('\x00')
        except Exception as error:
            logging.info(str(error))

        self.update_device_value_label()

    def to_bytes(self) -> bytes:

        byte_array = bytearray(self.on_file.name.encode())
        byte_array += bytearray(self.length_name - len(self.on_file.name))

        path_device = "ramfs:/" + self.on_file.name
        byte_array += bytearray(path_device.encode())
        byte_array += bytearray(self.length_path - len(path_device))

        return bytes(byte_array)

    def get_file_value(self):
        file_value = dict()

        file_value["length-name"] = self.length_name
        file_value["name"] = self.on_file.name
        file_value["length-path"] = self.length_path
        file_value["local-path"] = self.on_file.path

        return file_value

    def get_dynamic_memory_section(self) -> DynamicDataSet:

        try:
            file_desc = open(self.base_location + "\\" + self.on_file.path, "rb")
            data = file_desc.read()
        except Exception:
            raise DynamicSectionNotFoundException(self.base_location + "\\" + self.on_file.path)

        return WavefileDataSet(data)

    def update_device_value_label(self):
        # FIXME Not sure how to handle the modification status here...
        self.updated_device_value.emit(self.on_device.path, False)


class FileSystemDataHolder(WavefileDataHolder):
    def __init__(self, file_value: dict, base_location: str):
        super().__init__(file_value, base_location)

    def to_bytes(self) -> bytes:
        return bytes()

    def get_dynamic_memory_section(self) -> DynamicDataSet:
        if len(self.on_file.path) > 0:
            data = bytearray()
            try:
                file_desc = open(self.base_location + "\\" + self.on_file.path, "rb")
                data = file_desc.read()
            except Exception:
                raise DynamicSectionNotFoundException(self.base_location + "\\" + self.on_file.path)
            return FileSystemFileDataSet(data, self.path_device, self.length_path)
        else:
            pass


class StringDataHolder(DataHolder):

    def __init__(self, file_value: dict, is_read_only: bool = False):
        super().__init__()
        self.string = file_value["value"]
        self.maximum_length = file_value["maximum-length"]
        self.on_device_string = ""
        self.is_read_only = is_read_only

    def get_dynamic_memory_section(self) -> DynamicDataSet:
        pass

    def get_file_value(self):
        file_value = dict()

        file_value["maximum-length"] = self.maximum_length
        file_value["value"] = self.string

        return file_value

    def to_bytes(self) -> bytes:
        byte_array = bytearray(self.string.encode())
        byte_array += bytearray(self.maximum_length - len(self.string))
        return bytes(byte_array)

    def set_file_value(self, string):
        self.string = string
        self.update_on_device_string()

    def read_device_value(self, data: io.BytesIO):

        string = data.read(self.maximum_length)

        try:
            self.on_device_string = string.decode('ascii').strip().strip('\x00')
            self.update_on_device_string()
        except Exception as error:
            logging.info(str(error))

    def update_on_device_string(self):
        self.updated_device_value.emit(self.on_device_string, self.on_device_string != self.string)



import ctypes
import os
import wave

import tkinter.filedialog
import tkinter.messagebox


class Bos1921DataConverter:
    PCM16_TO_REFERENCE_VALUE = 18799
    MILLI_VOLT_TO_REFERENCE_VALUE = 18.34229

    def __init__(self, wave_file_path: wave, stabilization_value=0,
                 stabilization_duration_ms=5):

        # validate if the wave file path exist and throw an exception if the path is invalid
        if wave_file_path is None:
            raise Exception("The wave file path is null")

        # does the path exist?
        if not os.path.exists(wave_file_path):
            raise Exception("The wave file path does not exist")

        self.wave = wave_file_path
        self.stabilization_value = stabilization_value
        self.stabilization_duration_ms = stabilization_duration_ms

    def convert_to_bytes(self) -> bytearray:

        # open the wave file
        with wave.open(self.wave, "rb") as wave_file:
            # read the wave file
            wave_bytes = wave_file.readframes(wave_file.getnframes())

            # multiple each sample by Bos1921DataConverter.PCM16_TO_REFERENCE_VALUE
            wave_bytes = bytearray(wave_bytes)
            sample_bytes = bytearray()
            for i in range(0, len(wave_bytes), 2):
                sample = int.from_bytes(wave_bytes[i:i + 2], byteorder="little", signed=True)
                sample = (sample * 1000) / Bos1921DataConverter.PCM16_TO_REFERENCE_VALUE

                # rounding to the nearest integer
                sample = round(sample)
                sample = int(sample)

                sample = 0xFFF & sample
                sample = sample.to_bytes(2, byteorder="little", signed=False)

                # append the sample to the sample bytes
                sample_bytes += sample

            # return the bytes
            return sample_bytes


def apply_volume_and_factor_to_reference_data(_data_bytes, volume) -> int:
    for i in range(0, len(_data_bytes), 2):
        _sample = int.from_bytes(_data_bytes[i:i + 2], byteorder="little", signed=True)

        if _sample >= 0x800:
            _sample = ctypes.c_int16(_sample | 0xF000).value

        _sample = (_sample * volume * 1000) / (100 * 1000)
        _sample = int(_sample)

        _sample = _sample & 0xFFF

        _sample = _sample.to_bytes(2, byteorder="little", signed=True)
        _data_bytes[i:i + 2] = _sample

    return int(len(_data_bytes) / 2)


def convert_to_c_array_string(_data_bytes, byteorder="little") -> str:
    samples = list()
    # convert byte endianness from little to big where word size is 2 bytes
    for i in range(0, len(_data_bytes), 2):
        _sample = int.from_bytes(_data_bytes[i:i + 2], byteorder=byteorder)
        _sample = _sample.to_bytes(2, byteorder="big")
        samples[i:i + 2] = _sample

    # convert the bytes into a hex string
    data_string = "0x" + ", 0x".join("{:02x}".format(b) for b in samples)

    # return the string
    return data_string


def add_energy_management_for_press(_data_bytes: bytearray):
    vector = [0, -1, -3, -4, -7]

    for i in range(0, len(vector)):
        vector[i] &= 0x0FFF

    # add to the front of the byte array
    for i in range(-1 * len(vector), 0):
        _data_bytes[0:0] = bytearray(vector[i].to_bytes(2, byteorder="little", signed=False))

    return len(vector)


def add_stabilization(_data_bytes, stabilization_value_volt, stabilization_duration_ms, sampling_rate) -> int:
    # add release stabilization
    referenced_stab_value = int(stabilization_value_volt * Bos1921DataConverter.MILLI_VOLT_TO_REFERENCE_VALUE)

    nbr_samples = int((stabilization_duration_ms / 1000) * sampling_rate)
    referenced_stab_value &= 0xFFF
    stabilization_in_byte_array = referenced_stab_value.to_bytes(2, byteorder="little", signed=False)

    for i in range(0, nbr_samples):
        _data_bytes += stabilization_in_byte_array

    return nbr_samples


if __name__ == "__main__":
    # prompt user for wave file path in a windows dialog
    root = tkinter.Tk()
    root.withdraw()
    root.directory = tkinter.filedialog.askopenfilename(initialdir="/", title="Select file",
                                                        filetypes=(("wave files", "*.wav"), ("all files", "*.*")))

    # create a new instance of the data converter
    data_converter = Bos1921DataConverter(root.directory)
    data_bytes = data_converter.convert_to_bytes()

    waveform_nbr_sample = apply_volume_and_factor_to_reference_data(data_bytes, 66)
    stab_nbr_sample = add_stabilization(data_bytes, -0.150, 3, 8000)
    data_bytes += bytearray(0x0000.to_bytes(2, byteorder="little", signed=False))
    energy_management_nbr_sample = add_energy_management_for_press(data_bytes)

    print("waveform start index : %d stab index : %d Total Length %d", energy_management_nbr_sample,
          waveform_nbr_sample + energy_management_nbr_sample,
          waveform_nbr_sample + energy_management_nbr_sample + stab_nbr_sample)

    print(convert_to_c_array_string(data_bytes))

    # plot the data converted by the data converter
    import matplotlib.pyplot as plt

    samples = list()

    # convert the data sample to 2 bytes integers
    for i in range(0, len(data_bytes), 2):
        sample = int.from_bytes(data_bytes[i:i + 2], byteorder="little")
        if sample >= 0x800:
            sample = ctypes.c_int16(sample | 0xF000).value
        samples.append(sample)

    plt.plot(samples)
    plt.show()



import copy
import io
import json
import os

from PySide6.QtCore import QObject, Signal

from trackpad_configuration_manager.Bos1921Dataholder import Bos1921WavefileDataHolder
from trackpad_configuration_manager.dataHolder import NativeTypeDataHolder, WavefileDataHolder, DataHolder, \
    StringDataHolder, FileSystemDataHolder, BoolTypeDataHolder
from trackpad_configuration_manager.marshaller import ConfMarshaller
from trackpad_configuration_manager.probe.probe import Probe
from trackpad_configuration_manager.utils.execUtils import resource_path
from trackpad_configuration_manager.utils.hexFile import HexFile
from view.MessageDialog import ErrorMessageDialog


def remove_error_info(d):
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [remove_error_info(v) for v in d]
    return {k: remove_error_info(v) for k, v in d.items()
            if k not in {'errMsg', 'errCode'}}


class CRC(QObject):
    updated_crc_on_device = Signal(int, bool)
    updated_crc_on_file = Signal(int)

    def __init__(self):
        QObject.__init__(self)
        self._on_device = 0
        self._in_file = 0

    def set_crc_on_device(self, crc: int):
        self._on_device = crc
        self.updated_crc_on_device.emit(self._on_device, self._on_device == self._in_file)

    def set_crc_in_file(self, crc: int):
        self._in_file = crc
        self.updated_crc_on_file.emit(self._in_file)
        self.updated_crc_on_device.emit(self._on_device, self._on_device == self._in_file)


class BaseConfigFile:

    CRC_SIZE = 2

    def __init__(self, json_data: dict, file_path: str, name: str):
        self.json_data = json_data
        self.file_path = file_path
        self.device_address = json_data["target_address"]["value"]
        self.target_device = json_data["target_cpu"]["value"]

        if "target_space" in json_data:
            self.total_memory_available = json_data["target_space"]["value"]
        else:
            self.total_memory_available = 0

        self.name = name

        self._crc_on_device = 0
        self.trackpad_conf = ConfMarshaller(self.json_data)

        self.__add_on_device_value(self.json_data, io.BytesIO())
        self.crc = CRC()

    def read_device_data(self, probe: Probe) -> dict():
        binary, crc = self.trackpad_conf.get_binary_for_static_memory()
        self.crc.set_crc_in_file(crc)
        length = len(binary)

        data = probe.read_data(int(self.device_address, 16), length)
        byte_io = io.BytesIO(data)

        self.crc.set_crc_on_device(int.from_bytes(byte_io.read(BaseConfigFile.CRC_SIZE), byteorder="little"))

        try:
            self.__add_on_device_value(self.json_data, byte_io)
        except Exception:
            ErrorMessageDialog("Failed to parse data received from device").show()

    def get_file_data(self) -> dict():
        return self.json_data

    def get_memory_footprint(self) -> int:

        binary, crc = self.trackpad_conf.get_binary_for_static_memory()
        dynamic_binary = self.trackpad_conf.get_binaries_for_dynamic_memory()

        return len(binary) + len(dynamic_binary)

    def __get_binary(self) -> bytearray:
        binary, crc = self.trackpad_conf.get_binary_for_static_memory()
        dynamic_binary = self.trackpad_conf.get_binaries_for_dynamic_memory()

        return binary + dynamic_binary

    def save_on_device(self, probe: Probe):

        try:
            file_name = resource_path("conf.bin")
            file_binary = open(file_name, "wb")
            binary, crc = self.trackpad_conf.get_binary_for_static_memory()
            dynamic_binary = self.trackpad_conf.get_binaries_for_dynamic_memory()
            self.crc.set_crc_in_file(crc)
            file_binary.write(binary)

            if len(dynamic_binary) > 0:
                file_binary.write(dynamic_binary)

            file_binary.close()
            probe.write_data_from_file(file_name, int(self.device_address, 16))
        finally:
            os.remove(file_name)

    def get_hex_file(self) -> HexFile:
        return HexFile(self.__get_binary(), int(self.device_address, 0), self.name)

    def save_to_bin_file(self, file_name: str):
        with open(file_name, "wb") as file:
            file.write(self.__get_binary())

    def save_to_hex_file(self, hex_file_path: str):
        hex_file = self.get_hex_file()
        hex_file.generate_file(hex_file_path)

    @staticmethod
    def __handle_array_list_device_data(data_list: list, data: io.BytesIO):

        value_list = data_list["value"]
        for row in range(len(value_list)):
            for col in range(len(value_list[row])):

                if not isinstance(value_list[row][col], NativeTypeDataHolder):
                    value_list[row][col] = NativeTypeDataHolder(value_list[row][col], data_list["type"], False)

                value_list[row][col].read_device_value(data)

    def __add_on_device_value(self, d, data: io.BytesIO):

        if not isinstance(d, (dict, list)):
            return
        if isinstance(d, list):
            for v in d:
                self.__add_on_device_value(v, data)
        else:
            if "type" in d:
                match d["type"]:
                    case ("uint8_t" | "uint16_t" | "uint32_t" | "int8_t" | "int16_t" | "int32_t"):
                        # handle array list
                        if isinstance(d["value"], list):
                            self.__handle_array_list_device_data(d, data)
                        else:
                            if not isinstance(d["value"], NativeTypeDataHolder):
                                is_read_only = False
                                display_type = str()
                                choices = None

                                if "is_read_only" in d:
                                    is_read_only = d["is_read_only"]
                                if "display_type" in d:
                                    display_type = d["display_type"]
                                if "choices" in d:
                                    choices = d["choices"]

                                d["value"] = NativeTypeDataHolder(int(d["value"]), d["type"], is_read_only,
                                                                  choices, display_type)
                            d["value"].read_device_value(data)
                    case "string-wavefile-name-path":
                        if not isinstance(d["value"], WavefileDataHolder):
                            d["value"] = WavefileDataHolder(d["value"], self.file_path)
                        d["value"].read_device_value(data)
                    case "bos1921-referenced-wavefile":
                        if not isinstance(d["value"], Bos1921WavefileDataHolder):
                            d["value"] = Bos1921WavefileDataHolder(d["value"], self.file_path)
                        d["value"].read_device_value(data)
                    case "filesystem-file":
                        if not isinstance(d["value"], FileSystemDataHolder):
                            d["value"] = FileSystemDataHolder(d["value"], self.file_path)
                        d["value"].read_device_value(data)
                    case "string":
                        if not isinstance(d["value"], StringDataHolder):
                            is_read_only = False
                            if "is_read_only" in d:
                                is_read_only = d["is_read_only"]
                            d["value"] = StringDataHolder(d["value"], is_read_only)
                        d["value"].read_device_value(data)
                    case "bool":
                        if not isinstance(d["value"], BoolTypeDataHolder):
                            is_read_only = False
                            if "is_read_only" in d:
                                is_read_only = d["is_read_only"]
                            d["value"] = BoolTypeDataHolder(d["value"], is_read_only)
                        d["value"].read_device_value(data)

            for key, value in d.items():
                self.__add_on_device_value(value, data)

    def save_on_file(self, file_path: str = None):

        configuration_to_file = copy.copy(self.json_data)

        for key, element in configuration_to_file.items():
            configuration_to_file[key] = self._remove_data_holder(element)

        output_string = json.dumps(configuration_to_file, indent=4)

        if file_path is None:
            file_path = self.file_path

        with open(file_path, 'w') as outfile:
            outfile.write(output_string)

    def _remove_data_holder(self, element):
        if not isinstance(element, (dict, list)):
            return element
        if isinstance(element, list):
            list_json = list()
            for v in element:
                if isinstance(v, DataHolder):
                    list_json.append(v.get_file_value())
                else:
                    list_json.append(self._remove_data_holder(v))

            return list_json

        dict_json = dict()

        for k, v in element.items():
            if isinstance(v, DataHolder):
                dict_json[k] = v.get_file_value()
            else:
                dict_json[k] = self._remove_data_holder(v)

        return dict_json


class Calibration(BaseConfigFile):
    def __init__(self, json_data: dict, file_path: str):
        super(Calibration, self).__init__(json_data, file_path, "calibration")


class Configuration(BaseConfigFile):
    def __init__(self, json_data: dict, file_path: str):
        super(Configuration, self).__init__(json_data, file_path, "configuration-ux")



import wave

from trackpad_configuration_manager.boreas_data_converter.bos1921DataConverter import Bos1921DataConverter
from trackpad_configuration_manager.dataHolder import WavefileDataHolder, DynamicDataSet, \
    DynamicSectionNotFoundException, DynamicDataSetType
from trackpad_configuration_manager.utils.crc16 import compute_crc


class Bos1921FeedbackDataSet(DynamicDataSet):

    def __init__(self, data: bytearray, sampling_rate: int):
        self.sampling_rate_khz = int(sampling_rate / 1000)
        super().__init__(data, DynamicDataSetType.BOS1921_REFERENCED_FEEDBACK)

    def to_bytes(self) -> bytes:
        data = bytearray(self.type.value.to_bytes(length=1, byteorder="little", signed=False))  # data type
        data += bytearray(self.sampling_rate_khz.to_bytes(length=4, byteorder="little", signed=False))  # sampling rate
        data += bytearray(len(self.data).to_bytes(length=2, byteorder="little", signed=False))  # data length
        data += bytearray(self.data)  # data

        data += bytearray(compute_crc(data).to_bytes(length=2, byteorder="little", signed=False))  # crc

        return bytes(data)


class Bos1921WavefileDataHolder(WavefileDataHolder):
    def __init__(self, file_value: dict, base_location: str):
        super().__init__(file_value, base_location)
        self.__update_sampling_rate()

    def __update_sampling_rate(self):
        # open wavefile and read the sampling rate from the wavefile header
        self.sampling_rate = wave.open(self.base_location + "\\" + self.on_file.path, "rb").getframerate()

    def update_on_file(self, file_path: str):
        super().update_on_file(file_path)
        self.__update_sampling_rate()

    def get_dynamic_memory_section(self) -> DynamicDataSet:

        try:
            data_converter = Bos1921DataConverter(self.base_location + "\\" + self.on_file.path)
            data = data_converter.convert_to_bytes()
        except Exception:
            raise DynamicSectionNotFoundException(self.base_location + "\\" + self.on_file.path)

        return Bos1921FeedbackDataSet(data, self.sampling_rate)



import os

from intelhex import IntelHex


class HexFile:
    def __init__(self, data: bytearray, address_offset: int, name, align_length=4):
        file_name = "tmp-bin"

        padding_length = align_length - len(data) % align_length

        if padding_length > 0:
            data += bytearray(0x0 for _ in range(padding_length))

        file_binary = open(file_name, "wb")
        file_binary.write(data)
        file_binary.close()
        self.hFile = IntelHex()
        self.hFile.loadbin(file_name, address_offset)
        self.name = name

        os.remove(file_name)

    def generate_file(self, file_path: str):
        # create the file if it doesn't exist
        open(file_path, 'a').close()
        self.hFile.tofile(file_path, format='hex')



import crc


def compute_crc(data: bytearray) -> int:
    width = 16
    poly = 0x1189
    init_value = 0xFFFF
    final_xor_value = 0x00
    reverse_input = False
    reverse_output = False

    configuration = crc.Configuration(width, poly, init_value, final_xor_value, reverse_input, reverse_output)
    use_table = True
    crc_calculator = crc.CrcCalculator(configuration, use_table)

    return crc_calculator.calculate_checksum(data)



import json
import os
import sys

from trackpad_configuration_manager.configuration import Calibration, Configuration, BaseConfigFile


class HapticCalibrationProfileLevel:
    VOLUME_25 = "25%"
    VOLUME_50 = "50%"
    VOLUME_75 = "75%"
    VOLUME_100 = "100%"


class SensingCalibrationProfileLevel:
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class ConfigurationFactory:

    @staticmethod
    def create(file_name: str) -> BaseConfigFile:
        file = open(file_name)
        json_data = json.load(file)
        file.close()

        if {"file-type", "target_address", "target_cpu"}.issubset(json_data.keys()) is False:
            raise Exception("Invalid file format")

        match json_data["file-type"]["value"]:
            case "calibration":
                return Calibration(json_data, file_name)
            case "configuration":
                return Configuration(json_data, file_name)


class CalibrationDataSet(Calibration):
    CALIBRATION_STRING = "calibration"

    FEEDBACK_DICT = {"25%": 0, "50%": 1, "75%": 2, "100%": 3}

    def __init__(self, template_calibration_file_path: str):

        json_data = None

        with open(template_calibration_file_path, 'r') as file:
            json_data = json.load(file)

        if json_data is None:
            raise Exception("Error while loading json file")

        self.__validate_json_template(json_data)

        super(CalibrationDataSet, self).__init__(json_data, template_calibration_file_path)

        self.target_cpu = self.json_data["target_cpu"]["value"]

    def __validate_json_template(self, json_data):

        if "file-type" not in json_data:
            raise Exception("Invalid file format")

        if json_data["file-type"]["value"] != self.CALIBRATION_STRING:
            raise Exception("Wrong file type")

        if json_data["version"]["value"] != 3:
            raise Exception("Wrong version of json file")

        # not containing "target_address" key
        if "target_address" not in json_data:
            raise Exception("No target address in json file")

        if "target_cpu" not in json_data:
            raise Exception("No target cpu in json file")

    def get_nbr_zone_x(self):
        return self.json_data["configuration"]["division"]["nbWidthDivision"]

    def get_nbr_zone_y(self):
        return self.json_data["configuration"]["division"]["nbHeightDivision"]

    def update_sensing_profile(self, level: SensingCalibrationProfileLevel, sensing_matrix):

        # validate matrix is two dimensional
        if len(sensing_matrix) == 0 or len(sensing_matrix[0]) == 0:
            raise Exception("Invalid matrix size")

        if level == SensingCalibrationProfileLevel.LOW or level == SensingCalibrationProfileLevel.MEDIUM or \
                level == SensingCalibrationProfileLevel.HIGH:

            sensing_entries = self.json_data["configuration"]["value"]["sensing"]["value"]
            for index in range(len(sensing_entries)):
                if level in sensing_entries[index]:
                    matrix_ref_profile = sensing_entries[index][level]["value"]
                    self.__update_ref_matrix(matrix_ref_profile, sensing_matrix)
                    return

            raise Exception("Cannot update the sensing profile")
        else:
            raise Exception("Invalid sensing level")

    def update_feedback_profile(self, level: HapticCalibrationProfileLevel, feedback_matrix):

        # validate matrix is two dimensional
        if len(feedback_matrix) == 0 or len(feedback_matrix[0]) == 0:
            raise Exception("Invalid matrix size")

        if level == HapticCalibrationProfileLevel.VOLUME_25 or level == HapticCalibrationProfileLevel.VOLUME_50 or \
                level == HapticCalibrationProfileLevel.VOLUME_75 or \
                level == HapticCalibrationProfileLevel.VOLUME_100:

            feedback_entries = self.json_data["configuration"]["value"]["feedback"]["value"]
            for index in range(len(feedback_entries)):
                if level in feedback_entries[index]:
                    matrix_ref_profile = feedback_entries[index][level]["value"]
                    self.__update_ref_matrix(matrix_ref_profile, feedback_matrix)
                    return
        else:
            raise Exception("Invalid feedback level")

    def __update_ref_matrix(self, calibration_matrix_ref, new_matrix):

        # zero matrix
        for i in range(len(calibration_matrix_ref)):
            for j in range(len(calibration_matrix_ref[i])):
                calibration_matrix_ref[i][j].set_file_value_from_int(0)

        for i in range(len(new_matrix)):
            for j in range(len(new_matrix[i])):
                calibration_matrix_ref[i][j].set_file_value_from_int(new_matrix[i][j])

    def save_output_json_file(self, output_json_file: str, output_directory=""):

        output_calibration_file = os.path.join(output_directory, output_json_file)

        # determine if application is a script file or frozen exe
        if getattr(sys, 'frozen', False):
            output_calibration_file = os.path.join(os.path.dirname(sys.executable), output_calibration_file)

        output_calibration_file = os.path.join(output_directory, output_calibration_file)

        self.save_on_file(output_calibration_file)

    def save_to_hex_file(self, hex_file_path: str):
        hex_file = self.get_hex_file()
        hex_file.generate_file(hex_file_path)



import re

from pyocd.core.helpers import ConnectHelper
from pyocd.flash.file_programmer import FileProgrammer
import time

from trackpad_configuration_manager.cmsis_package import CMSISPackage


class Probe:
    OPS_DELAY_SEC = 0.1
    WORD_LENGTH = 4
    MAX_WORD_PER_READ = 64000

    def __init__(self, unique_id: str, target_id: str, blocking=True, delay_between_ops=False):

        CMSISPackage.load_package()

        self.delay_between_ops = delay_between_ops
        self.unique_id = unique_id
        self.session = ConnectHelper.session_with_chosen_probe(unique_id=unique_id, target_override=target_id,
                                                               blocking=blocking,
                                                               options={"connect_mode": "under-reset",
                                                                        "reset_type": "hw",
                                                                        "hide_programming_progress": True})

        if not self.session:
            raise Exception("Session cannot be established with " + unique_id)

        self.target = self.session.board.target

    def __del__(self):
        self.session.close()

    @staticmethod
    def __nbr_words_for(byte_length: int) -> int:
        length = int(byte_length / Probe.WORD_LENGTH)
        if byte_length % Probe.WORD_LENGTH:
            length += 1
        return length

    @staticmethod
    def __convert_to_byte_array(data_list: list) -> bytearray:
        data = bytearray()
        for word in data_list:
            data += bytearray(word.to_bytes(length=4, byteorder="little", signed=False))

        return data

    @staticmethod
    def scan_for_regex(data_string: str, regex: str) -> str:
        version = str()

        matches = re.finditer(regex, data_string, re.MULTILINE)

        for matchNum, match in enumerate(matches):
            version = match.group()

        return version

    def __read_memory_as_ascii_data(self, address: int, word_length: int) -> str:

        data_list = self.target.read_memory_block32(address, word_length)

        data = self.__convert_to_byte_array(data_list)
        return data.decode('ascii', errors="ignore")

    FULL_VERSION_REGEX = r"\d+\.\d+\.\d+-*\w*\+\d+.\b[0-9a-f]{5,40}\b-*\w*"

    def scan_for_version(self) -> str:

        try:
            self.session.open()
            self.target.halt()

            if self.delay_between_ops is True:
                time.sleep(Probe.OPS_DELAY_SEC)

            flash = self.target.memory_map.get_boot_memory()
            length = self.__nbr_words_for(flash.end - flash.start)

            nbr_words_per_scan = Probe.MAX_WORD_PER_READ if length >= Probe.MAX_WORD_PER_READ else length

            nbr_iteration = int(length / nbr_words_per_scan)

            version = str()

            for i in range(nbr_iteration):
                data = self.__read_memory_as_ascii_data(flash.start + i * nbr_words_per_scan,
                                                        nbr_words_per_scan)

                # detect the string version using regex where the format is x.y.z-TAG+w.hash
                version = self.scan_for_regex(data, self.FULL_VERSION_REGEX)

                if len(version):
                    break
        finally:
            self.target.reset_and_halt()
            self.target.resume()
            self.session.close()

        return version

    def write_data_from_file(self, file_name: str, address: int):

        try:
            self.session.open()
            self.target.halt()

            if self.delay_between_ops is True:
                time.sleep(Probe.OPS_DELAY_SEC)

            programmer = FileProgrammer(self.session)
            programmer.program(file_name, "bin", base_address=address)

        finally:
            self.target.reset_and_halt()
            self.target.resume()
            self.session.close()

    def read_data(self, address: int, length: int) -> bytearray:

        try:
            self.session.open()
            self.target.halt()

            word_length = 4
            length_to_read = int(length / word_length)
            if length % word_length:
                length_to_read += 1

            if self.delay_between_ops is True:
                time.sleep(Probe.OPS_DELAY_SEC)

            data_list = self.target.read_memory_block32(address, length_to_read)
        finally:
            self.target.reset_and_halt()
            self.target.resume()
            self.session.close()

        return self.__convert_to_byte_array(data_list)



import sys
import setuptools
from setuptools.command.install import install

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


class InstallCmsisPack(install):
    def run(self):
        print("sys.argv:", sys.argv)
        print("Running custom installation install logic...")
        # Download the files
        self.download_cmsis_packs()

        # Continue with the default installation process
        install.run(self)

    def download_cmsis_packs(self):
        import os
        try:
            import requests
        except ImportError:
            # If requests is not available, provide a warning message
            print("Warning: 'requests' library not found. File downloads will be skipped.")

        # URLs and download paths
        urls = [
            'https://www.keil.com/pack/Keil.STM32F0xx_DFP.2.1.1.pack',
            'http://www.keil.com/pack/Keil.STM32G0xx_DFP.1.4.0.pack',
            'https://www.keil.com/pack/Keil.STM32U5xx_DFP.2.0.0.pack'
        ]
        download_paths = [
            'cmsis-pack/Keil.STM32F0xx_DFP.2.1.1.pack',
            'cmsis-pack/Keil.STM32G0xx_DFP.1.4.0.pack',
            'cmsis-pack/Keil.STM32U5xx_DFP.2.0.0.pack'
        ]

        # Download the files only if requests is available
        if 'requests' in locals():
            for url, download_path in zip(urls, download_paths):
                # download to the subfolder download_path
                abs_download_path = os.path.join(self.install_lib, "trackpad_configuration_manager", download_path)

                os.makedirs(os.path.dirname(abs_download_path), exist_ok=True)

                print("Downloading file from {} to {}".format(url, abs_download_path))
                response = requests.get(url)
                with open(abs_download_path, 'wb') as f:
                    f.write(response.content)


setuptools.setup(
    name="trackpad_configuration_manager",
    version="2.1.1",
    author="Pascal-Frédéric St-Laurent",
    author_email="pfstlaurent@boreas.ca",
    description="Trackpad Configurator Library that allows to read/write configuration/calibration files (json format) "
                "from/to a device",
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    url="https://github.com/BoreasTechnologies/trackpad-configurator-plus-plus.git",
    packages=[
        'trackpad_configuration_manager',
        'trackpad_configuration_manager.probe',
        'trackpad_configuration_manager.utils',
        'trackpad_configuration_manager.boreas_data_converter'],
    classifiers=[
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.11",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=[
        "pyocd==0.34.3",
        "cmsis-pack-manager==0.5.2",
        "intelhex==2.3.0",
        "PySide6==6.5.2",
        "crc==1.3.0",
        "requests==2.31.0",
    ],
    cmdclass={
        'install': InstallCmsisPack,
    },
    include_package_data=True,
    python_requires='>=3.11',
)



from datetime import datetime

from trackpad_configuration_manager.probe.probe import Probe
from trackpad_configuration_manager.utils.calibration_data_set import CalibrationDataSet, \
    SensingCalibrationProfileLevel, HapticCalibrationProfileLevel

# example: How to load a calibration file template, update and save it on a device via the ST-Link probe
if __name__ == "__main__":
    calibration_data_set = CalibrationDataSet("template/calibration_Gen2.json")

    # update the sensing profile for medium sensitivity
    # These value are computed using the calibration methodology provided by Boreas separately

    #                   Top of trackpad
    sensing_matrix = [[1, -2, 0],  # index [0,0], [0,1], [0,2]
                      [2, 10, 1],  # index [1,0], [1,1], [1,2]
                      [3, 12, 3]]  # index [2,0], [2,1], [2,2]
    #               Bottom of trackpad

    calibration_data_set.update_sensing_profile(SensingCalibrationProfileLevel.MEDIUM, sensing_matrix)

    # update the haptic profile for 50% volume
    # These value are computed using the calibration methodology provided by Boreas separately

    #                   Top of trackpad
    feedback_matrix = [[10, 5, 2],  # index [0,0], [0,1], [0,2]
                       [4, 5, 6],   # index [1,0], [1,1], [1,2]
                       [7, 8, 9]]   # index [2,0], [2,1], [2,2]
    #               Bottom of trackpad

    calibration_data_set.update_feedback_profile(HapticCalibrationProfileLevel.VOLUME_50, feedback_matrix)

    # save the updated json file with timestamp in the file name
    json_output_name = "calibration_Gen2_" + datetime.now().strftime("%Y%m%d_%H%M%S") + ".json"
    calibration_data_set.save_output_json_file(json_output_name)

    # save on the device via ST-Link probe. The unique id is the serial number of the ST-Link probe.
    unique_id = "002000343137510439383538"
    probe = Probe(unique_id, calibration_data_set.target_cpu, blocking=False, delay_between_ops=True)

    calibration_data_set.save_on_device(probe)

    # save the calibration data set in hex and bin format
    calibration_data_set.save_to_hex_file("calibration_Gen2.hex")
    calibration_data_set.save_to_bin_file("calibration_Gen2.bin")
